"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.CustomAttributeRegistry = void 0;
const helpers_1 = require("./helpers");
class CustomAttributeRegistry {
    constructor(parent) {
        this.parent = parent;
        this._attrMap = new Map();
        this._elementMap = new Map();
        if (!parent) {
            throw new Error('Must be given a parent element');
        }
        this.observe();
    }
    define(attrName, Constructor) {
        this._attrMap.set(attrName.toLowerCase(), Constructor);
        this.upgradeAttribute(attrName);
    }
    get(element, attrName) {
        const map = this._elementMap.get(element);
        if (!map)
            return;
        return map.get(attrName.toLowerCase());
    }
    getConstructor(attrName) {
        return this._attrMap.get(attrName.toLowerCase());
    }
    observe() {
        var _a, _b;
        this.observer = new MutationObserver((mutations) => {
            for (const mutation of mutations) {
                if (mutation.type === 'attributes') {
                    const attr = this.getConstructor(mutation.attributeName);
                    if (attr) {
                        this.found(mutation.attributeName, mutation.target, mutation.oldValue);
                    }
                }
                else {
                    for (const node of mutation.removedNodes) {
                        this.downgrade(node);
                    }
                    for (const node of mutation.addedNodes) {
                        this.upgradeElement(node);
                    }
                }
            }
        });
        this.observer.observe((_b = (_a = this.parent) === null || _a === void 0 ? void 0 : _a['shadowRoot']) !== null && _b !== void 0 ? _b : this.parent, {
            childList: true,
            subtree: true,
            attributes: true,
            attributeOldValue: true,
        });
    }
    unsubscribe(force) {
        var _a;
        if (force) {
            (_a = this.observer) === null || _a === void 0 ? void 0 : _a.disconnect();
        }
        const values = [...this._elementMap.values()];
        for (const elModifiers of values) {
            const modifiers = [...elModifiers.values()];
            for (const modifier of modifiers) {
                modifier.OnDestroy();
                if (modifier.observer) {
                    modifier.observer.disconnect();
                }
            }
            elModifiers.clear();
        }
        this._elementMap.clear();
    }
    upgradeAttribute(attrName, doc) {
        const parent = doc || this.parent;
        const matches = parent.querySelectorAll('[' + attrName + ']');
        for (const match of [...matches]) {
            this.found(attrName, match);
        }
    }
    upgradeElement(element) {
        if (element.nodeType !== 1) {
            return;
        }
        for (const attr of element.attributes) {
            if (this.getConstructor(attr.name)) {
                this.found(attr.name, element);
            }
        }
        for (const [attr] of this._attrMap) {
            this.upgradeAttribute(attr, element);
        }
    }
    downgrade(element) {
        const map = this._elementMap.get(element);
        if (!map) {
            return;
        }
        for (const [, instance] of map) {
            if (instance.OnDestroy) {
                instance.OnDestroy();
            }
        }
        this._elementMap.delete(element);
    }
    found(attributeName, el, oldVal) {
        var _a, _b;
        let map = this._elementMap.get(el);
        if (!map) {
            map = new Map();
            this._elementMap.set(el, map);
        }
        let modifier = map.get(attributeName);
        const attribute = el.getAttribute(attributeName);
        if (!modifier) {
            const Modifier = this.getConstructor(attributeName);
            modifier = new Modifier();
            if ((_b = (_a = Modifier.options) === null || _a === void 0 ? void 0 : _a.observedAttributes) === null || _b === void 0 ? void 0 : _b.length) {
                for (const observedAttribute of Modifier.options.observedAttributes) {
                    (0, helpers_1.observe)(modifier, observedAttribute);
                }
            }
            modifier.element = el;
            modifier.selector = attributeName;
            modifier.value = attribute || modifier.value;
            modifier.parent = this.parent;
            if (modifier.OnInit) {
                modifier.OnInit();
            }
            if (Modifier.options.observe) {
                modifier.observer = new MutationObserver((records) => modifier.OnChange(records));
                modifier.observer.observe(modifier.element, Modifier.options.observe);
            }
            map.set(attributeName, modifier);
            return;
        }
        if (attribute == null && !!modifier.value) {
            modifier.value = attribute;
            if (modifier.OnDestroy) {
                modifier.OnDestroy();
            }
            if (modifier.observer) {
                modifier.observer.disconnect();
            }
            map.delete(attributeName);
            return;
        }
        if (attribute !== modifier.value) {
            modifier.value = attribute;
            if (modifier.OnUpdate) {
                modifier.OnUpdate(oldVal, attribute);
            }
            return;
        }
    }
}
exports.CustomAttributeRegistry = CustomAttributeRegistry;
//# sourceMappingURL=custom-registry.js.map